from kgtools.preprocessing import Pipeline
from kgtools.func import parallel
from kgtools.saver import Saver

if __name__ == "__main__":
    conf = {
        "cleaner": {
            "rules": [{"type": "root_node", "attr": "class", "value": "devsite-article-body"}]
        }
    }
    pipeline = Pipeline(conf=conf)
    data = Saver.load("testdata/guide.bin")
    docs, sentences = pipeline.process(data)

    Saver.dump(docs, "testdata/docs.bin")
    Saver.dump(sentences, "testdata/sentences.bin")

    with Path("testdata/sentences.txt").open("w", encoding="utf-8") as f:
        f.write("\n".join([str(sent) for sent in sentences]))
    with Path("testdata/tokens.txt").open("w", encoding="utf-8") as f:
        f.write("\n".join([str(sent) + "\n" + sent.tokenized_text + "\n" + sent.lemma_text + "\n" for sent in sentences]))
    with Path("testdata/vocab.txt").open("w", encoding="utf-8") as f:
        f.write("\n".join(sorted(vocab.words, key=lambda x: x)))
    print(vocab.get_emb("-PRON-"))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from kgtools.annotation import Lazy
from kgtools.type.span import Span


class Sentence:
    def __init__(self, text, docs=None, tokens=None, nps=None):
        self.text = text
        self.docs = docs
        self.tokens = tokens
        self.nps = set() if nps is None else nps

    def __str__(self):
        return self.text

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __len__(self):
        return len(self.tokens)

    def __iter__(self):
        return iter(self.tokens)

    def __add__(self, other):
        if self == other:
            sent = Sentence(self.text, self.docs | other.docs, self.tokens)
            for doc in sent.docs:
                doc.sents[doc.sent2index[sent]] = sent
            return sent
        else:
            print("The two sentences must have the same 'text'")
            return self

    def __iadd__(self, other):
        if self == other:
            self.docs |= other.docs
            return self
        else:
            print("The two sentences must have the same 'text'")
            return self

    def add_nps(self, *pairs):
        self.nps.update({Span(self, *pair) for pair in pairs})

    def find_spans(self, *spans, is_lemma=True):
        words = [token.lemma if is_lemma else token.text for token in self.tokens]
        group_dict = {}
        for span in spans:
            length = len(span.split())
            if length in group_dict:
                group_dict[length].add(span)
            else:
                group_dict[length] = {span}
        group_dict = dict(sorted(group_dict.items(), key=lambda x: x[0], reverse=True))
        index = 0
        result_dict = {}
        while index < len(self):
            for length, group in group_dict.items():
                span = " ".join(words[index:index + length])
                if span in group:
                    result_dict[span] = index
                    index += length
                    break
            else:
                index += 1
        result = [(span, result_dict.get(span, -1)) for span in spans]
        return result

    @Lazy
    def tokenized_text(self):
        return " ".join([token.text for token in self.tokens])

    @Lazy
    def lemma_text(self):
        return " ".join([token.lemma for token in self.tokens])

    @Lazy
    def emb(self):
        return sum([token.emb for token in self.tokens]) / len(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from kgtools.annotation import Lazy

from kgtools.type.vocab import Vocab


class Token:
    def __init__(self, text, lemma, vocab=Vocab(), pos=None, dep=None, ner=None):
        self.text = text
        self.lemma = lemma
        self.vocab = vocab
        self.lemma_first = vocab.lemma_first
        self.vocab.add(lemma if self.lemma_first else text)
        self.pos = pos
        self.dep = dep
        self.ner = ner

    def __str__(self):
        return self.lemma if self.lemma_first else self.text

    def __hash__(self):
        return hash(str(self))

    @Lazy
    def emb(self):
        return self.vocab.get_emb(str(self))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from nltk.tokenize import sent_tokenize as st, word_tokenize as wt
import re
import spacy
from type import Vocab, Doc, Sentence, Token




class Tokenizer:

    __name__ = "Tokenizer"

    PUNC_TABLE = {ord(zh): ord(en) for zh, en in zip('‘’“”…，。！？【】（）％＃＠＆：',
                                                     '\'\'"".,.!?[]()%#@&:')}

    # MARK_TABLE = {
    #     "__CODE__": ""
    # }

    def __init__(self, vocab=Vocab(), code_patterns=None):
        self.vocab = vocab
        self.patterns = [
            re.compile(r'([a-zA-Z_][a-zA-Z\d_]*)(\.\1)+'),
            re.compile(r'[a-zA-Z_]([a-z][A-Z])'),
            re.compile(r'[a-zA-Z_][a-zA-Z\d_]*\([a-zA-Z\d_]*\)(\.[a-zA-Z_][a-zA-Z\d_]*\([a-zA-Z\d_]*\))'),
            re.compile(r'')
        ]
        code_patterns = code_patterns if code_patterns is not None else []
        self.patterns.extend(code_patterns)
        
        self.nlp = spacy.load('en')
        hyphen_re = re.compile(r"[A-Za-z\d]+-[A-Za-z\d]+|'[a-z]+|''")
        prefix_re = spacy.util.compile_prefix_regex(self.nlp.Defaults.prefixes)
        infix_re = spacy.util.compile_infix_regex(self.nlp.Defaults.infixes)
        suffix_re = spacy.util.compile_suffix_regex(self.nlp.Defaults.suffixes)
        self.nlp.tokenizer = spacy.tokenizer.Tokenizer(self.nlp.vocab, prefix_search=prefix_re.search, infix_finditer=infix_re.finditer,
                               suffix_search=suffix_re.search, token_match=hyphen_re.match)

    def sent_tokenize(self, text):
        text = text.translate(Tokenizer.PUNC_TABLE)
        text = re.sub(r'\s+', ' ', text).strip()
        text = re.sub(r'({[^{}]*?)(\?)([^{}]*?})', r'\1__?__\3', text)
        text = re.sub(r'(\[[^\[\]]*?)(\?)([^\[\]]*?\])', r'\1__?__\3', text)
        text = re.sub(r'(\([^()]*?)(\?)([^()]*?\))', r'\1__?__\3', text)
        text = re.sub(r'(\<[^<>]*?)(\?)([^<>]*?\>)', r'\1__?__\3', text)
        text = re.sub(r'("[^"]*?)(\?)([^"]*?")', r'\1__?__\3', text)

        text = re.sub(r'({[^{}]*?)(!)([^{}]*?})', r'\1__!__\3', text)
        text = re.sub(r'(\[[^\[\]]*?)(!)([^\[\]]*?\])', r'\1__!__\3', text)
        text = re.sub(r'(\([^()]*?)(!)([^()]*?\))', r'\1__!__\3', text)
        text = re.sub(r'(\<[^<>]*?)(!)([^<>]*?\>)', r'\1__!__\3', text)
        text = re.sub(r'("[^"]*?)(!)([^"]*?")', r'\1__!__\3', text)

        text = re.sub(r'({[^{}]*?)(\.)([^{}]*?})', r'\1__.__\3', text)
        text = re.sub(r'(\[[^\[\]]*?)(\.)([^\[\]]*?\])', r'\1__.__\3', text)
        text = re.sub(r'(\([^()]*?)(\.)([^()]*?\))', r'\1__.__\3', text)
        text = re.sub(r'(\<[^<>]*?)(\.)([^<>]*?\>)', r'\1__.__\3', text)
        text = re.sub(r'("[^"]*?)(\.)([^"]*?")', r'\1__.__\3', text)

        text = text.replace("e.g.", "__eg__")
        text = text.replace("E.g.", "__eg__")
        text = text.replace("E.G.", "__eg__")
        text = text.replace("i.e.", "__ie__")
        text = text.replace("I.e.", "__ie__")
        text = text.replace("I.E.", "__ie__")
        sentences = []
        for sent in st(text):
            if self.__pre_check(sent):
                sent_text = sent.replace("__eg__", "e.g.").replace("__ie__", "i.e.").replace("__?__", "?").replace("__!__", "!").replace("__.__", ".")
                sent_text = re.sub(r'^(-CODE- |-TAB- |-IMG- |-URL- )(.*)', r'\2', sent_text)
                sent_text = re.sub(r'^(\()(.*)(\))$', r'\2', sent_text)
                sent_text = re.sub(r'^(\[)(.*)(\])$', r'\2', sent_text)
                sent_text = re.sub(r'^({)(.*)(})$', r'\2', sent_text)
                words = sent_text.split()
                if re.search(r'^[^A-Z]', words[0]) is not None and words[1] in {"A", "An", "The", "This", "That", "You", "We"} and re.search(r'^[^A-Z]', words[2]) is None:
                    sent_text = " ".join(words[1:])
                sent_text = sent_text.strip()
                if self.__post_check(sent_text):
                    sentences.append(Sentence(sent_text))
        # text = re.sub(r'\n(.+?[^.?!])\n([A-Z])', r'\n\n\2', text)
        # text = re.sub(r'\s+', " ", text.strip())
        # text = re.sub(r'([?!.]+) ', r'\1\n', text)
        # sentences = set(text.split("\n"))
        return sentences

    def __pre_check(self, sentence):
        if len(sentence) == 0 or not (5 <= len(sentence.split()) <= 200):
            return False
        # check chinese
        if any(["\u4e00" <= ch <= "\u9fff" for ch in sentence]):
            return False
        return True

    def __post_check(self, sentence):
        if re.search(r'^[0-9a-zA-Z"\'<(]', sentence) is None:
            return False
        if sentence.count('[') != sentence.count(']'):
            return False
        if sentence.count('(') != sentence.count(')'):
            return False
        if sentence.count('{') != sentence.count('}'):
            return False
        if sentence.count('"') != sentence.count('"'):
            return False
        if sentence.count(':') > 3 or sentence.count('=') > 3 or sentence.count('[') > 3 or sentence.count('{') > 3:
            return False
        return True

    def word_tokenize_nltk(self, sentence):
        sentence = sentence.replace("e.g.", "__eg__").replace("E.g.", "__eg__").replace("E.G.", "__eg__").replace("i.e.", "__ie__").replace("I.e.", "__ie__").replace("I.E.", "__ie__")
        sentence = re.sub(r'([a-zA-Z ])\.([a-zA-Z ])', r'\1 . \2', sentence)
        sentence = re.sub(r'\s+', ' ', sentence)

        tokens = wt(sentence)
        tokens = [t.replace("__eg__", "e.g.").replace("__ie__", "i.e.") for t in tokens]
        return tokens

    def word_tokenize_spacy(self, sentence):
        # sentence = sentence.replace("__CODE__", "CODE$").replace("__IMG__", "IMG$").replace("__TAB__", "TAB$").replace("__URL__", "URL$").replace("__QUOTE__", "QUOTE$")
        tokens = [Token(t.text, t.lemma_, vocab=self.vocab, pos=t.pos_, dep=t.dep_) for t in self.nlp(sentence)]
        # for token in tokens:
        #     token.text = token.text.replace("CODE$", "__CODE__").replace("IMG$", "__IMG__").replace("TAB$", "__TAB__").replace("URL$", "__URL__").replace("QUOTE$", "__QUOTE__")
        #     token.lemma = token.lemma.replace("code$", "__code__").replace("img$", "__img__").replace("tab$", "__tab__").replace("url$", "__url__").replace("quote$", "__quote__")
        return tokens

    def tokenize(self, rawdocs):
        docs = set()
        sent2sent = {}
        for rawdoc in rawdocs:
            sents = []
            for text in rawdoc:
                sents.extend(self.sent_tokenize(text))
            if len(sents) > 0:
                docs.add(Doc(rawdoc.url, sents))
                for sent in sents:
                    if sent in sent2sent:
                        new_sent = sent + sent2sent.pop(sent)
                        sent2sent[new_sent] = new_sent
                    else:
                        sent2sent[sent] = sent
        sentences = set(sent2sent.values())
        for sent in sentences:
            tokens = self.word_tokenize_nltk(sent.text)
            sent.tokens = self.word_tokenize_spacy(" ".join(tokens))
            sent.docs = {doc.url for doc in sent.docs}
        return docs, sentences, self.vocab

    # def __handle_sent(self, rawdocs):
    #     docs = set()
    #     sent2sent = {}
    #     for rawdoc in rawdocs:
    #         sents = []
    #         for text in rawdoc:
    #             sents.extend(self.sent_tokenize(text))
    #         if len(sents) > 0:
    #             docs.add(Doc(rawdoc.url, sents))
    #             for sent in sents:
    #                 if sent in sent2sent:
    #                     new_sent = sent + sent2sent.pop(sent)
    #                     sent2sent[new_sent] = new_sent
    #                 else:
    #                     sent2sent[sent] = sent
    #     return docs, [sent2sent]

    # def __handle_word(self, sents):
    #     sentence_texts = []
    #     for sent in sents:
    #         tokens = self.word_tokenize_nltk(sent.text)
    #         sentence_texts.append((sent, " ".join(tokens)))
    #     return sentence_texts

    # def tokenize_parallel(self, rawdocs):

    #     docs, sent2sent_list = parallel(self.__handle_sent, list(rawdocs))
    #     sent2sent = sent2sent_list.pop(0)
    #     for _ in range(len(sent2sent_list)):
    #         cur = sent2sent_list.pop(0)
    #         for sent in cur.keys():
    #             if sent in sent2sent:
    #                 new_sent = sent + sent2sent.pop(sent)
    #                 sent2sent[new_sent] = new_sent
    #             else:
    #                 sent2sent[sent] = sent

    #     sentences = set(sent2sent.values())

    #     pairs = parallel(self.__handle_word, list(sentences))
    #     for sent, sent_text in pairs:
    #         sent.tokens = self.word_tokenize_spacy(sent_text)
    #         sent.docs = {doc.url for doc in sent.docs}
    #     return docs, sentences

    def find_code(self, sentence):
        '''find code elements

        [description]

        Arguments:
            sentence {[type]} -- [description]
        '''
        for pattern in self.patterns:
            pass


if __name__ == "__main__":
    t = Tokenizer()
    tokens = t.word_tokenize("Those are the language-specific tokenizer(exceptions), for example:__CODE__")
    for token in tokens:
        print(token.text, token.lemma, token.pos, token.dep)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# from kgtools.nlp.tokenizer import CompoundTokenizer

# tokenizer = CompoundTokenizer()
# sent = tokenizer.word_tokenize("The word-frequency returns the sequence element's frequency over training corpus.")
# print(sent)
# for token in sent:
#     print(token)

from kgtools.type.vocab import Vocab



#!/usr/bin/env python
# -*- coding: utf-8 -*-

class Span:
    def __init__(self, sentence, start, end):
        self.sentence = sentence
        self.start = start
        self.end = end

    def __str__(self):
        return " ".join([str(t) for t in self.sentence.tokens[self.start:self.end]])

    def __add__(self, other):
        if self.sentence == other.sentence and self.end == other.start:
            self.end = other.end
        return self


#!/usr/bin/env python
# -*- coding: utf-8 -*-

from abc import ABCMeta, abstractmethod
from typing import List
import re
from bs4 import BeautifulSoup

from kgtools.annotation import Parallel, TimeLog
from kgtools.symbol import HTML


class HTMLParser:
    __name__ = "HTMLParser"

    class Node:
        def __init__(self, key, value):
            assert key in {"name", "class_", "id"}, "The parameter 'key' must be in {'name', 'class_', 'id'}"
            self.key = key
            self.value = value

    def __init__(self, entry_nodes: List[Node]=None, filter_nodes: List[Node]=None):
        self.entry_nodes = entry_nodes if entry_nodes is not None else []
        self.filter_nodes = filter_nodes if filter_nodes is not None else []

    def parse(self, html):
        body = BeautifulSoup(html, "lxml").body

        # remove useless elements
        scripts = body.findAll("script")
        [script.extract() for script in scripts]
        noscripts = body.findAll("noscript")
        [noscript.extract() for noscript in noscripts]
        navs = body.findAll(class_=re.compile(r'.*(nav|Nav|footer|Footer).*'))
        [nav.extract() for nav in navs]
        footers = body.findAll("footer")
        [footer.extract() for footer in footers]

        for node in self.filter_nodes:
            filtered = body.findAll(**{node.key: node.value})
            [n.extract() for n in filtered]

        entries = []
        for node in self.entry_nodes:
            entries.extend(body.findAll(**{node.key: node.value}))
        if len(entries) == 0:
            entries = [body]

        texts = set()
        for entry in entries:
            for li in entry.findAll("li"):
                string = li.get_text().strip()
                if len(string) > 0 and string[-1] not in set(".?!:;,"):
                    string = string + "."
                li.clear()
                li.append(string)
            for h in entry.findAll(re.compile(r'h[1-6]')):
                string = h.get_text().strip()
                if len(string) > 0 and string[-1] not in set(".?!:;,"):
                    string = string + "."
                h.clear()
                h.append(string)
            for p in entry.findAll("p"):
                string = p.get_text().strip()
                if len(string) > 0 and string[-1] not in set(".?!:;,"):
                    string = string + "."
                p.clear()
                p.append(string)

            for table in entry.findAll("table"):
                table.clear()
                table.append(f"{HTML.TAB}")
            for img in entry.findAll("img"):
                if not img.get("alt") or len(img["alt"]) == 0:
                    img_alt = f"{HTML.IMG}"
                else:
                    img_alt = img["alt"]
                img.insert_after(img_alt)
            for code in entry.findAll("code"):
                string = code.get_text().strip()
                if len(string.split()) > 5 or len(string) > 50:
                    string = f"{HTML.CODE}"
                code.clear()
                code.append(string)

            for pre in entry.findAll("pre"):
                pre.clear()
                pre.append(f"{HTML.PRE}.")
            for pre in entry.findAll("blockquote"):
                pre.clear()
                pre.append(f"{HTML.QUOTE}.")
            text = entry.get_text()
            text = text.strip() + " "
            text = re.sub(r'(https?://.*?)([^a-zA-Z0-9/]?\s)', r"%s\2" % HTML.TAB, text)
            text = re.sub(r'\s+', ' ', text).strip()
            if len(text) > 0:
                if text[-1] not in set(".?!:;,"):
                    text = text + "."
                texts.add(text)
        return texts

    @TimeLog
    @Parallel()
    def process(self, html_list):
        docs = set()
        for html in html_list:
            texts = self.parse(html)
            docs.update(texts)
        return docs


class JavadocParser(HTMLParser):
    __name__ = "JavadocParser"

    def __init__(self, **cfg):
        super(self.__class__, self).__init__(**cfg)

    def parse(self, html):
        body = BeautifulSoup(html, "lxml").body
        strings = []
        for div in body.select(".block"):
            string = div.get_text().strip()
            string = re.sub(r'\s+', ' ', string)
            if len(string) > 0 and string[-1] not in set(".?!"):
                string = string + "."
            strings.append(string)
        return strings


if __name__ == "__main__":
    html_parser = HTMLParser()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

__version__ = '0.1.11'

setup(
    name='kgtools',
    version=__version__,
    description='The tools for KG Team of Fudan SE LAB',
    url='https://github.com/FudanSELab/KG-Tools',
    author='Chong Wang',
    author_email='18212010030@fudan.edu.cn',
    license='MIT',
    packages=find_packages(exclude=("docs", "test")),
    zip_safe=False,
    python_requires='>=3',
    install_requires=[
        'bs4',
        'lxml',
        'spacy',
        'nltk',
        'gensim',
        'numpy',
        'pathos',
        'dill',
        'aiohttp'
    ],
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import threading
import numpy as np


class Vocab:
    __thread_lock = threading.Lock()
    # _process_lock = multiprocessing.Lock()

    def __new__(cls, *args, **kwargs):
        if not hasattr(Vocab, "_instance"):
            with Vocab.__thread_lock:
                if not hasattr(Vocab, "_instance"):
                    Vocab._instance = object.__new__(cls)
        return Vocab._instance

    def __init__(self, lemma_first=True, stopwords=None, emb_size=100):
        self.words = set()
        self.embedding = {}
        self.stopwords = stopwords
        self.emb_size = emb_size

        self.lemma_first = lemma_first

        self.ZERO = np.array([0.] * self.emb_size)

    @classmethod
    def new_instance(cls, *args, **kwargs):
        instance = object.__new__(cls)
        instance.__init__(*args, **kwargs)
        return instance

    def get_emb(self, word):
        return self.embedding.get(word, self.ZERO)

    def add(self, word):
        self.words.add(word)

    def __len__(self):
        return len(self.words)

    def __getitem__(self, key):
        return self.get_emb(key)

    def __add__(self, other):
        vocab = Vocab(self.lemma_first, self.stopwords, self.emb_size)
        vocab.words = self.words | other.words
        vocab.embedding = dict(set(self.embedding.items()) | (other.embedding.items()))
        vocab.stopwords = self.stopwords
        if self.stopwords is not None:
            if other.stopwords is not None:
                self.stopwords.update(other.stopwords)
        else:
            self.stopwords = other.stopwords
        return vocab

    def __iadd__(self, other):
        self.words |= other.words
        self.embedding.update(other.embedding)
        if self.stopwords is not None:
            if other.stopwords is not None:
                self.stopwords.update(other.stopwords)
        else:
            self.stopwords = other.stopwords
        return self
